"""Module to validate indicator source data before uploading to the public COVIDcast API.

This file defines the functions that are made public by the module. As the
module is intended to be executed though the main method, these are primarily
for testing.
"""
