# -*- coding: utf-8 -*-
"""Call the function params_run when executed.

This file indicates that calling the module (`python -m delphi_utils`) will
call the function `params_run` found within the utils.py file. There should be
no need to change this template.
"""

from .utils import params_run # pragma: no cover

params_run() # pragma: no cover
