# -*- coding: utf-8 -*-
"""Call the function run_module when executed.

This file indicates that running the module (`python -m delphi_validator`) will
call the function `run_module` found within the run.py file.
"""

from .run import run_module

run_module()
