# -*- coding: utf-8 -*-
"""Module to flagging interesting or unusual data points.

This file defines the functions that are made public by the module. As the
module is intended to be executed though the main method, these are primarily
for testing.
"""
from __future__ import absolute_import
from .constants import HTML_LINK
